#include <stdio.h>
#include <string.h>
#include <openssl/evp.h>
#include "crypto_kem.h"

long long cpucycles(void)
{
  unsigned long long result;
  asm volatile(".byte 15;.byte 49;shlq $32,%%rdx;orq %%rdx,%%rax"
    : "=a" (result) ::  "%rdx");
  return result;
}

static int aes256ctr_xor(
  unsigned char *out,
  const unsigned char *in,
  unsigned long long inlen,
  const unsigned char *n,
  const unsigned char *k
)
{
  EVP_CIPHER_CTX *x;
  int ok;
  int outl = 0;

  x = EVP_CIPHER_CTX_new();
  if (!x) return -111;

  ok = EVP_EncryptInit_ex(x,EVP_aes_256_ctr(),0,k,n);
  if (ok == 1) ok = EVP_CIPHER_CTX_set_padding(x, 0);
  if (ok == 1) ok = EVP_EncryptUpdate(x, out, &outl, in, inlen);
  if (ok == 1) ok = EVP_EncryptFinal_ex(x, out, &outl);

  EVP_CIPHER_CTX_free(x);
  return ok == 1 ? 0 : -111;
}

static int crypto_stream(
  unsigned char *out,
  unsigned long long outlen,
  const unsigned char *n,
  const unsigned char *k
)
{
  unsigned char temp[outlen];
  memset(temp, 0, outlen);
  return aes256ctr_xor(out, temp, outlen, n, k);
}

#define crypto_rng_OUTPUTBYTES 736

static const unsigned char nonce[16] = {0};

static int crypto_rng(
        unsigned char *r, /* random output */
        unsigned char *n, /* new key */
  const unsigned char *g  /* old key */
)
{
  unsigned char x[32 + crypto_rng_OUTPUTBYTES];
  crypto_stream(x,sizeof x,nonce,g);
  memcpy(n,x,32);
  memcpy(r,x + 32,crypto_rng_OUTPUTBYTES);
  return 0;
}

static unsigned char g[32];
static unsigned char r[crypto_rng_OUTPUTBYTES];
static unsigned long long pos = crypto_rng_OUTPUTBYTES;

void randombytes(unsigned char *x,unsigned long long xlen)
{
  while (xlen > 0) {
    unsigned long long ready;
    
    if (pos == crypto_rng_OUTPUTBYTES) {
      while (xlen > crypto_rng_OUTPUTBYTES) {
        crypto_rng(x,g,g);
        x += crypto_rng_OUTPUTBYTES;
        xlen -= crypto_rng_OUTPUTBYTES;
      }
      if (xlen == 0) return;
      
      crypto_rng(r,g,g);
      pos = 0;
    }
    
    ready = crypto_rng_OUTPUTBYTES - pos;
    if (xlen <= ready) ready = xlen;
    memcpy(x,r + pos,ready);
    memset(r + pos,0,ready);
    x += ready;
    xlen -= ready;
    pos += ready;
  }
}

unsigned char pk[crypto_kem_PUBLICKEYBYTES];
unsigned char sk[crypto_kem_SECRETKEYBYTES];
unsigned char c[crypto_kem_CIPHERTEXTBYTES];
unsigned char k[crypto_kem_BYTES];

#define TIMINGS 21
long long t[TIMINGS];

int main()
{
  int i;

  for (i = 0;i < TIMINGS;++i) {
    t[i] = cpucycles();
  }
  printf("nothing");
  for (i = 1;i < TIMINGS;++i) printf(" %lld",t[i]-t[i-1]);
  printf("\n");

  for (i = 0;i < TIMINGS;++i) {
    crypto_kem_keypair(pk,sk);
    t[i] = cpucycles();
  }
  printf("keypair");
  for (i = 1;i < TIMINGS;++i) printf(" %lld",t[i]-t[i-1]);
  printf("\n");

  for (i = 0;i < TIMINGS;++i) {
    crypto_kem_enc(c,k,pk);
    t[i] = cpucycles();
  }
  printf("enc");
  for (i = 1;i < TIMINGS;++i) printf(" %lld",t[i]-t[i-1]);
  printf("\n");

  for (i = 0;i < TIMINGS;++i) {
    crypto_kem_dec(k,c,sk);
    t[i] = cpucycles();
  }
  printf("dec");
  for (i = 1;i < TIMINGS;++i) printf(" %lld",t[i]-t[i-1]);
  printf("\n");

  return 0;
}
